import torch

from helios import losses


@losses.LOSS_REGISTRY.register
class SampleLoss(losses.WeightedLoss):
    def __init__(self):
        super().__init__(loss_weight=0.5)

    def _eval(self, x: torch.Tensor) -> torch.Tensor:
        return x


class TestLosses:
    def test_registry(self, check_registry) -> None:
        check_registry(losses.LOSS_REGISTRY, ["SampleLoss"])

    def test_create(self, check_create_function) -> None:
        check_create_function(losses.LOSS_REGISTRY, losses.create_loss)

    def test_weighted_loss(self) -> None:
        loss = losses.create_loss("SampleLoss")
        assert isinstance(loss, SampleLoss)

        x = torch.tensor(10)
        y = loss(x)
        assert y == (x * 0.5)



import typing

import numpy as np
import numpy.typing as npt
import torch
from torch import nn

from helios import core

from .functional import (
    calculate_f1_torch,
    calculate_mae_torch,
    calculate_mAP,
    calculate_precision_torch,
    calculate_psnr,
    calculate_psnr_torch,
    calculate_recall_torch,
    calculate_ssim,
    calculate_ssim_torch,
)

METRICS_REGISTRY = core.Registry("metrics")


def create_metric(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> nn.Module:
    """
    Create the metric function for the given type.

    Args:
        type_name (str): the type of the loss to create.
        args: positional arguments to pass into the metric.
        kwargs: keyword arguments to pass into the metric.

    Returns:
        Callable: the metric function
    """
    return METRICS_REGISTRY.get(type_name)(*args, **kwargs)


@METRICS_REGISTRY.register
class CalculatePSNR(nn.Module):
    """
    Calculate PSNR (Peak Signal-to-Noise Ratio).

    Implementation follows: https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio
    Note that the input_order is only needed if you plan to evaluate Numpy images. It can
    be left as default otherwise.

    Args:
        crop_border (int): Cropped pixels in each edge of an image. These pixels are not
        involved in the calculation.
        input_order (str): Whether the input order is 'HWC' or 'CHW'.
        test_y_channel (bool): Test on Y channel of YCbCr. Default:
    """

    def __init__(
        self, crop_border: int, input_order: str = "HWC", test_y_channel: bool = False
    ):
        """Construct the PSNR metric."""
        super().__init__()

        self._crop_border = crop_border
        self._input_order = input_order
        self._test_y_channel = test_y_channel

    def forward(
        self, img: npt.NDArray | torch.Tensor, img2: npt.NDArray | torch.Tensor
    ) -> float:
        """
        Calculate the PSNR metric.

        Args:
            img (np.ndarray): Images with range [0, 255].
            img2 (np.ndarray): Images with range [0, 255].

        Returns:
            float: PSNR value.
        """
        if isinstance(img, torch.Tensor) and isinstance(img2, torch.Tensor):
            return calculate_psnr_torch(
                img, img2, self._crop_border, self._test_y_channel
            )
        assert isinstance(img, np.ndarray) and isinstance(img2, np.ndarray)
        return calculate_psnr(
            img, img2, self._crop_border, self._input_order, self._test_y_channel
        )


@METRICS_REGISTRY.register
class CalculateSSIM(nn.Module):
    """
    Calculate SSIM (structural similarity).

    Implementation follows: 'Image quality assesment: From error visibility to structural
    similarity'. Results are identical to those of the official MATLAB code in
    https://ece.uwaterloo.ca/~z70wang/research/ssim/.
    For three-channel images, SSIM is calculated for each channel and then
    averaged.

    Args:
        crop_border (int): Cropped pixels in each edge of an image. These pixels are not
        involved in the calculation.
        input_order (str): Whether the input order is 'HWC' or 'CHW'.
        test_y_channel (bool): Test on Y channel of YCbCr.

    Returns:
        float: SSIM.
    """

    def __init__(
        self, crop_border: int, input_order: str = "HWC", test_y_channel: bool = False
    ):
        """Construct the SSIM metric."""
        super().__init__()

        self._crop_border = crop_border
        self._input_order = input_order
        self._test_y_channel = test_y_channel

    def forward(
        self, img: npt.NDArray | torch.Tensor, img2: npt.NDArray | torch.Tensor
    ) -> float:
        """
        Calculate the SSIM metric.

        Args:
            img (np.ndarray): Images with range [0, 255].
            img2 (np.ndarray): Images with range [0, 255].

        Returns:
            float: PSNR value.
        """
        if isinstance(img, torch.Tensor) and isinstance(img2, torch.Tensor):
            return calculate_ssim_torch(
                img, img2, self._crop_border, self._test_y_channel
            )
        assert isinstance(img, np.ndarray) and isinstance(img2, np.ndarray)
        return calculate_ssim(
            img, img2, self._crop_border, self._input_order, self._test_y_channel
        )


@METRICS_REGISTRY.register
class CalculateMAP(nn.Module):
    """
    Calculate the mAP (Mean Average Precision).

    Implementation follows:
    https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision
    """

    def forward(self, targs: npt.NDArray, preds: npt.NDArray) -> float:
        """
        Calculate the mAP (Mean Average Precision).

        Args:
            targs (np.ndarray): target (inferred) labels in range [0, 1].
            preds (np.ndarray): predicate labels in range [0, 1].

        Returns:
            float: the mAP score
        """
        return calculate_mAP(targs, preds)


@METRICS_REGISTRY.register
class CalculateMAE(nn.Module):
    """
    Calculate the mAP (Mean Average Precision).

    Implementation follows:
    https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision
    """

    def forward(self, pred: torch.Tensor, gt: torch.Tensor) -> float:
        """
        Calculate the mAP (Mean Average Precision).

        Args:
            targs (np.ndarray): target (inferred) labels in range [0, 1].
            preds (np.ndarray): predicate labels in range [0, 1].

        Returns:
            float: the mAP score
        """
        return calculate_mae_torch(pred, gt)


@METRICS_REGISTRY.register
class CalculatePrecision(nn.Module):
    """
    Compute the Precision score.

    Implementation follows:  https://en.wikipedia.org/wiki/Precision_and_recall
    """

    def forward(self, pred: torch.Tensor, gt: torch.Tensor) -> float:
        """
        Compute the Precision score.

        Args:
            pred (torch.Tensor): predicate (inferred) images in range [0, 255]
            gt (torch.Tensor): ground-truth images in range [0, 255]

        Returns:
            float: the precision score.
        """
        return calculate_precision_torch(pred, gt)


@METRICS_REGISTRY.register
class CalculateRecall(nn.Module):
    """
    Compute the Recall score.

    Implementation follows:  https://en.wikipedia.org/wiki/Precision_and_recall
    """

    def forward(self, pred: torch.Tensor, gt: torch.Tensor) -> float:
        """
        Compute the Recall score.

        Args:
            pred (torch.Tensor): predicate (inferred) images in range [0, 255]
            gt (torch.Tensor): ground-truth images in range [0, 255]

        Returns:
            float: the recall score.
        """
        return calculate_recall_torch(pred, gt)


@METRICS_REGISTRY.register
class CalculateF1(nn.Module):
    """
    Compute the F1 score.

    Implementation follows: https://en.wikipedia.org/wiki/F-score
    """

    def forward(self, pred: torch.Tensor, gt: torch.Tensor) -> float:
        """
        Compute the F1 score.

        Args:
            pred (torch.Tensor): predicate (inferred) images in range [0, 255]
            gt (torch.Tensor): ground-truth images in range [0, 255]

        Returns:
            float: the F1 score.
        """
        return calculate_f1_torch(pred, gt)



import os
import pathlib
import typing

import torch
import torch.nn.functional as F
import torchvision
import torchvision.transforms.v2 as T
from torch import nn

import helios.core as hlc
import helios.data as hld
import helios.model as hlm
import helios.optim as hlo
import helios.trainer as hlt
from helios.core import logging


class CIFARDataModule(hld.DataModule):
    """
    Example datamodule class built with CIFAR10.

    Here you can see how to setup a datamodule that requires data to be downloaded using
    torchvision's CIFAR10 dataset. The code is adapted from PyTorch's "Training a
    Classifier" tutorial:
    https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html

    Args:
        root (pathlib.Path): the root where the data will be downloaded to.
    """

    def __init__(self, root: pathlib.Path) -> None:
        """Construct the datamodule."""
        super().__init__()
        self._root = root / "data"

    def prepare_data(self) -> None:
        """Download all necessary data."""
        torchvision.datasets.CIFAR10(root=self._root, train=True, download=True)
        torchvision.datasets.CIFAR10(root=self._root, train=False, download=True)

    def setup(self) -> None:
        """Create the datasets."""
        # Use the ToTensor transform from Helios to automate the conversion from images to
        # tensors.
        transforms = T.Compose(
            [hld.transforms.ToTensor(), T.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
        )
        params = hld.DataLoaderParams()
        params.batch_size = 4
        params.shuffle = True
        params.num_workers = 2
        params.drop_last = True
        self._train_dataset = self._create_dataset(
            torchvision.datasets.CIFAR10(
                root=self._root, train=True, download=False, transform=transforms
            ),
            params,
        )

        # The dataloader params are copied when the dataset is created, so we can safely
        # change the options for the validation dataset without interfering with the ones
        # for training.
        params.drop_last = False
        params.shuffle = False
        self._valid_dataset = self._create_dataset(
            torchvision.datasets.CIFAR10(
                root=self._root, train=False, download=False, transform=transforms
            ),
            params,
        )


class Net(nn.Module):
    """
    Example image classifier.

    The code is taken from PyTorch's Training a Classifier tutorial:
    https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html
    """

    def __init__(self):
        """Create the classifier."""
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the label for the given image."""
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)  # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class ClassifierModel(hlm.Model):
    """
    Example model class for training the classifier.

    Here you can see how to setup the model class and some of the basic functionality that
    is available. The code is adapted from PyTorch's "Training a Classifier" tutorial:
    https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html
    """

    def __init__(self) -> None:
        """Create the model."""
        super().__init__("classifier")

    def setup(self, fast_init: bool = False) -> None:
        """Create the network, loss, and optimizer."""
        # Note that when we create the network and loss function, we immediately move them
        # to the current device, which has been set by the trainer.
        self._net = Net().to(self.device)
        self._criterion = nn.CrossEntropyLoss().to(self.device)

        # Note that SGD is shipped as part of the default optimizers from Helios, so we
        # can directly request it from create_optimizer instead of building it ourselves.
        self._optimizer = hlo.create_optimizer(
            "SGD", self._net.parameters(), lr=0.001, momentum=0.9
        )

    def load_state_dict(
        self, state_dict: dict[str, typing.Any], fast_init: bool = False
    ) -> None:
        """Restore the model from a saved checkpoint."""
        # Note that we don't have to re-map the weights ourselves. They have already been
        # re-mapped for us by the trainer when it loaded the checkpoint.
        self._net.load_state_dict(state_dict["net"])
        self._criterion.load_state_dict(state_dict["criterion"])
        self._optimizer.load_state_dict(state_dict["optimizer"])

    def state_dict(self) -> dict[str, typing.Any]:
        """Return the state dict of the model."""
        return {
            "net": self._net.state_dict(),
            "criterion": self._criterion.state_dict(),
            "optimizer": self._optimizer.state_dict(),
        }

    def train(self) -> None:
        """Set the model to train."""
        # If we had more networks, we would shift them to training mode here.
        self._net.train()

    def on_training_start(self) -> None:
        """Perform steps before training starts."""
        tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())

        x = torch.randn((1, 3, 32, 32)).to(self.device)
        tb_logger.add_graph(self._net, x)

    def train_step(self, batch: typing.Any, state: hlt.TrainingState) -> None:
        """Forward and backward training passes."""
        # Due to the simplicity of the code, we do both the forward and backward passes in
        # the training step, but you could also split it between the train_step and
        # on_training_batch_end if your setup is more complex.
        inputs, labels = batch
        inputs = inputs.to(self.device)
        labels = labels.to(self.device)

        self._optimizer.zero_grad()

        outputs = self._net(inputs)
        loss = self._criterion(outputs, labels)
        loss.backward()
        self._optimizer.step()

        # Note that we save the value of the loss function to the _loss_items dictionary.
        # This allows the model to automatically gather the losses for us if we are
        # training in distributed mode.
        self._loss_items["loss"] = loss

    def on_training_batch_end(
        self,
        state: hlt.TrainingState,
        should_log: bool = False,
    ) -> None:
        """
        Perform steps after the training batch.

        Args:
            state (pyt.TrainingState): the current training state.
            should_log (bool): if true, then writing to the log should be performed.
        """
        # If we were training in distributed mode, calling the base Model's function will
        # automatically gather all loss values saved to self._loss_items.
        super().on_training_batch_end(state, should_log)

        # This flag is set to true whenever the number of iterations is a multiple of the
        # logging frequency we set when creating the trainer.
        if should_log:
            root_logger = logging.get_root_logger()
            tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())

            loss_val = self._loss_items["loss"]

            root_logger.info(
                f"[{state.global_epoch + 1}, {state.global_iteration:5d}] "
                f"loss: {loss_val:.3f}, "
                f"running loss: {loss_val / state.running_iter:.3f} "
                f"avg time: {state.average_iter_time:.2f}s"
            )
            tb_logger.add_scalar("train/loss", loss_val, state.global_iteration)
            tb_logger.add_scalar(
                "train/running loss",
                loss_val / state.running_iter,
                state.global_iteration,
            )

    def on_training_end(self) -> None:
        """Perform steps after training ends."""
        # For our example, we're going to save the hyper-params to the tensorboard log so
        # we can compare with other runs.
        # Notice that self._val_scores is active because we validate every epoch. If your
        # validation frequency is different, you may need to alter this code.
        total = self._val_scores["total"]
        correct = self._val_scores["correct"]
        accuracy = 100 * correct // total
        writer = hlc.get_from_optional(logging.get_tensorboard_writer())
        writer.add_hparams(
            {"lr": 0.001, "momentum": 0.9, "epochs": 2},
            {"hparam/accuracy": accuracy, "hparam/loss": self._loss_items["loss"].item()},
        )

    def eval(self) -> None:
        """Set the model to eval."""
        self._net.eval()

    def on_validation_start(self, validation_cycle: int) -> None:
        """Perform steps on validation start."""
        # The base function will automatically clear the validation scores for us.
        super().on_validation_start(validation_cycle)

        # If you need to add further data to the table, you can do so here. In our case,
        # we're going to add the total number of labels seen and how many of those were
        # correct so we can compute the accuracy metric.
        self._val_scores["total"] = 0
        self._val_scores["correct"] = 0

    def valid_step(self, batch: typing.Any, step: int) -> None:
        """Perform the validation step."""
        images, labels = batch
        images = images.to(self.device)
        labels = labels.to(self.device)

        outputs = self._net(images)

        _, predicted = torch.max(outputs.data, 1)
        self._val_scores["total"] += labels.size(0)
        self._val_scores["correct"] += (predicted == labels).sum().item()

    def on_validation_end(self, validation_cycle: int) -> None:
        """Perform steps after validation ends."""
        root_logger = logging.get_root_logger()
        tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())

        # Grab the validation scores and compute the accuracy metric so we can log it. If
        # we were in distributed mode, you would need to gather the values here.
        total = self._val_scores["total"]
        correct = self._val_scores["correct"]
        accuracy = 100 * correct // total

        root_logger.info(f"[Validation {validation_cycle}] accuracy: {accuracy}")
        tb_logger.add_scalar("val", accuracy, validation_cycle)


if __name__ == "__main__":
    # Set the CUBLAS workspace setting to allow determinism to be used in CUDA >= 10.2.
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"

    datamodule = CIFARDataModule(pathlib.Path.cwd())
    model = ClassifierModel()

    trainer = hlt.Trainer(
        run_name="cifar10",
        train_unit=hlt.TrainingUnit.EPOCH,
        total_steps=2,
        valid_frequency=1,
        chkpt_frequency=1,
        print_frequency=10,
        enable_tensorboard=True,
        enable_file_logging=True,
        enable_progress_bar=True,
        enable_deterministic=True,
        chkpt_root=pathlib.Path.cwd() / "chkpt",
        log_path=pathlib.Path.cwd() / "logs",
        run_path=pathlib.Path.cwd() / "runs",
    )

    trainer.fit(model, datamodule)



import pathlib

import numpy as np
import onnx
import onnxruntime  # type: ignore[import-untyped]
import torch
import torch.onnx
from torch import nn


def export_to_onnx(
    net: nn.Module,
    net_args: torch.Tensor,
    out_path: pathlib.Path,
    validate_output: bool = False,
    rtol: float = 1e-3,
    atol: float = 1e-5,
    **kwargs,
) -> None:
    """
    Export the given network to ONNX format.

    By default, the resulting onnx network will be validated through ONNX to ensure it's
    valid. If you wish to validate the traced outputs to ensure they're the same, set
    validate_output to True and change rtol/atol as needed.

    Args:
        net (nn.Module): the network to convert.
        net_args (torch.Tensor): the input tensor for tracing.
        out_path (pathlib.Path): the path to save the exported network to.
        validate_output (bool): if True, validation is performed to ensure correctness.
        rtol (float): relative tolerance threshold.
        atol (float): absolute tolerance threshold.
        kwargs (dict): additional keyword arguments to torch.onnx.export.
    """
    net.eval()
    with torch.no_grad():
        out = net(net_args)

    torch.onnx.export(net, net_args, out_path, **kwargs)

    onnx_model = onnx.load(out_path)
    onnx.checker.check_model(onnx_model)

    if validate_output:
        ort_session = onnxruntime.InferenceSession(
            out_path, providers=["CPUExecutionProvider"]
        )

        def to_numpy(tensor: torch.Tensor):
            if tensor.requires_grad:
                return tensor.detach().cpu().numpy()
            return tensor.cpu().numpy()

        ort_inputs = {ort_session.get_inputs()[0].name: to_numpy(net_args)}
        ort_outs = ort_session.run(None, ort_inputs)
        np.testing.assert_allclose(to_numpy(out), ort_outs[0], rtol=rtol, atol=atol)



from helios import optim


class TestOptimizers:
    def test_registry(self, check_registry) -> None:
        check_registry(optim.OPTIMIZER_REGISTRY, ["Adam", "AdamW", "SGD"])

    def test_create(self, check_create_function) -> None:
        check_create_function(optim.OPTIMIZER_REGISTRY, optim.create_optimizer)



from helios import metrics


class TestMetrics:
    def test_registry(self, check_registry) -> None:
        check_registry(
            metrics.METRICS_REGISTRY,
            [
                "CalculatePSNR",
                "CalculateSSIM",
                "CalculateMAP",
                "CalculateMAE",
                "CalculatePrecision",
                "CalculateRecall",
                "CalculateF1",
            ],
        )

    def test_create(self, check_create_function) -> None:
        check_create_function(metrics.METRICS_REGISTRY, metrics.create_metric)



from __future__ import annotations

import math
import typing

import torch
import torch.utils.data as tud

from helios import core

SAMPLER_REGISTRY = core.Registry("sampler")


def create_sampler(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> ResumableSamplerType:
    """
    Create a sampler of the given type.

    This uses the SAMPLER_REGISTRY to look-up sampler types, so ensure your samplers have
    been registered before using this function.

    Args:
        type_name (str): the type of the transform to create.
        args: positional arguments to pass into the sampler.
        kwargs: keyword arguments to pass into the sampler.

    Returns:
        ResumableSamplerType: the constructed sampler.
    """
    return SAMPLER_REGISTRY.get(type_name)(*args, **kwargs)


class ResumableSampler(tud.Sampler):
    """
    Base class for samplers that are resumable.

    In this context, a resumable sampler is defined as a sampler that, when training stops
    and is later resumed, guarantees that the order in which the batches are returned is
    always the same and that they will continue from where the previous run left off.
    """

    def __init__(self, batch_size: int) -> None:
        """Create the sampler."""
        super().__init__()

        self._start_iter: int = 0
        self._epoch: int = 0
        self._batch_size = batch_size

    def _adjust_to_start_iter(self, indices: list[int]) -> list[int]:
        assert self._batch_size > 0

        start_index = self._start_iter * self._batch_size
        return indices[start_index:]

    @property
    def start_iter(self) -> int:
        """The starting iteration for the sampler."""
        return self._start_iter

    @start_iter.setter
    def start_iter(self, ite: int) -> None:
        self._start_iter = ite

    def set_epoch(self, epoch: int) -> None:
        """Set the current epoch for seeding."""
        self._epoch = epoch


@SAMPLER_REGISTRY.register
class ResumableRandomSampler(ResumableSampler):
    """
    Random sampler with resumable state.

    This allows training to stop and resume while guaranteeing that the order in which the
    batches will be returned stays consistent. It is effectively a replacement to the
    default RandomSampler from PyTorch.

    Args:
        data_source (Sized): the dataset to sample from.
        seed (int): the seed to use for setting up the random generator.
        batch_size (int): the number of samples per batch.
    """

    def __init__(
        self, data_source: typing.Sized, seed: int = 0, batch_size: int = 1
    ) -> None:
        """Create the sampler."""
        super().__init__(batch_size)

        self._data_source = data_source
        self._seed = seed
        self._num_samples = len(self._data_source)

    def __len__(self) -> int:
        """Return the length of the dataset."""
        return self._num_samples

    def __iter__(self) -> typing.Iterator[int]:
        """Retrieve the index of the next sample."""
        g = torch.Generator()
        g.manual_seed(self._epoch + self._seed)
        shuffling = torch.randperm(self._num_samples, generator=g)
        indices = shuffling.tolist()

        assert len(indices) == self._num_samples
        indices = self._adjust_to_start_iter(indices)
        return iter(indices)


@SAMPLER_REGISTRY.register
class ResumableSequentialSampler(ResumableSampler):
    """
    Sequential sampler with resumable state.

    This allows training to stop and resume while guaranteeing that the order in which the
    batches will be returned stays consistent. It is effectively a replacement to the
    default SequentialSampler from PyTorch.

    Args:
        data_source (Sized): the dataset to sample from.
        batch_size (int): the number of samples per batch.
    """

    def __init__(self, data_source: typing.Sized, batch_size: int = 1):
        """Create the sampler."""
        super().__init__(batch_size)

        self._data_source = data_source
        self._num_samples = len(data_source)
        self._indices = list(range(self._num_samples))

    def __len__(self) -> int:
        """Return the length of the dataset."""
        return self._num_samples

    def __iter__(self) -> typing.Iterator[int]:
        """Retrieve the index of the next sample."""
        assert self._batch_size > 0
        indices = self._indices
        indices = self._adjust_to_start_iter(indices)
        return iter(indices)


@SAMPLER_REGISTRY.register
class ResumableDistributedSampler(tud.DistributedSampler):
    """
    Distributed sampler with resumable state.

    This allows training to stop and resume while guaranteeing that the order in which the
    batches will be returned stays consistent. It is effectively a replacement to the
    default DistributedSampler from PyTorch.

    Args:
        dataset (torch.utils.data.Dataset): the dataset to sample from.
        num_replicas (int | None): number of processes for distributed training.
        rank (int | None): rank of the current process.
        shuffle (bool): if True, shuffle the indices.
        seed (int): random seed used to shuffle the sampler.
        drop_last (bool): If true, then drop the final sample to make it even across
        replicas.
        batch_size (int): the number of samples per batch.
    """

    def __init__(
        self,
        dataset: tud.Dataset,
        num_replicas: int | None = None,
        rank: int | None = None,
        shuffle: bool = True,
        seed: int = 0,
        drop_last: bool = False,
        batch_size: int = 1,
    ):
        """Create the sampler."""
        super().__init__(dataset, num_replicas, rank, shuffle, seed, drop_last)
        self._batch_size = batch_size
        self._start_iter = 0

    @property
    def start_iter(self) -> int:
        """The starting iteration for the sampler."""
        return self._start_iter

    @start_iter.setter
    def start_iter(self, ite: int) -> None:
        self._start_iter = ite

    def __iter__(self) -> typing.Iterator[int]:
        """Retrieve the index of the next sample."""
        # The code for this function was adapted from PyTorch's implementation of
        # DistributedSampler. The original license from PyTorch can be viewed here:
        # https://github.com/pytorch/pytorch/blob/main/LICENSE
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()  # type: ignore[arg-type]
        else:
            indices = list(range(len(self.dataset)))  # type: ignore[arg-type]

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[
                    :padding_size
                ]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[: self.total_size]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        # Trim off samples that we have already seen if we're restarting.
        assert self._batch_size > 0
        start_index = self._start_iter * self._batch_size
        indices = indices[start_index:]

        return iter(indices)


ResumableSamplerType = ResumableSampler | ResumableDistributedSampler



from helios import scheduler


class TestSchedulers:
    def test_registry(self, check_registry) -> None:
        check_registry(
            scheduler.SCHEDULER_REGISTRY,
            [
                "MultiStepLR",
                "CosineAnnealingLR",
                "CosineAnnealingRestartLR",
                "MultiStepRestartLR",
            ],
        )

    def test_create(self, check_create_function) -> None:
        check_create_function(scheduler.SCHEDULER_REGISTRY, scheduler.create_scheduler)



import pathlib

import cv2
import numpy as np
import numpy.typing as npt
import PIL
import torch


def load_image(path: pathlib.Path, out_fmt: str = "") -> npt.NDArray:
    """
    Load the given image and convert it to a numpy array.

    out_fmt is a format string that can be passed in to PIL.Image.convert. Please see the
    documentation for accepted strings. If no string is passed, the image will be
    converted to RGB format.

    Args:
        path (pathlib.Path): the path to the image to load.
        out_fmt (str): the format to convert the loaded image to.

    Returns:
        np.ndarray: the loaded image.
    """
    with path.open(mode="rb") as infile:
        img = PIL.Image.open(infile)
        out = img.convert(out_fmt) if out_fmt != "" else img.convert("RGB")
        return np.array(out)


def tensor_to_numpy(tens: torch.Tensor, as_float: bool = False) -> npt.NDArray:
    """
    Convert the given tensor to a numpy array.

    Args:
        tens (torch.Tensor): the tensor to convert in the range [0, 1]
        as_float (bool): whether to leave the output as float or convert to int.

    Returns:
        np.ndarray: the converted array.
    """
    as_np = tens.squeeze().float().clamp_(0, 1).cpu().detach().numpy()
    if as_np.ndim == 3:
        as_np = np.transpose(as_np, (1, 2, 0))

    if not as_float:
        as_np = np.uint8((as_np * 255.0).round())

    return as_np


def show_tensor(tens: torch.Tensor, title: str = "debug window") -> None:
    """
    Display the image held by the tensor. Useful for debugging purposes.

    Args:
        tens (torch.Tensor): the image tensor to display in range [0, 1].
        title (str): the title of the displayed window.
    """
    img = tensor_to_numpy(tens)
    img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
    cv2.imshow(title, img)
    cv2.waitKey(0)


def show_tensors(tens: torch.Tensor) -> None:
    """
    Show batches of tensors.

    Args:
        tens (torch.Tensor): the batch of tensors to display in range [0, 1].
    """
    if len(tens.shape) == 3:
        show_tensor(tens)
        return

    for b in range(tens.shape[0]):
        show_tensor(tens[b], title=f"tensor[{b}]")


def convert_to_hwc(img: npt.NDArray, input_order: str = "HWC") -> npt.NDArray:
    """
    Change the order of the input image channels so the result is in (h, w, c) order.

    If the input image is a single-channel image, then the return is (h, w, 1).

    Args:
        img (np.ndarray): input image.
        input_order (str): the order of the channels of the input image. Must be one of
        'HWC' or 'CHW'

    Returns:
        np.ndarray: the shuffled image.
    """
    assert input_order in ("HWC", "CHW")

    if len(img.shape) == 2:
        img = img[..., None]
    if input_order == "CHW":
        img = img.transpose(1, 2, 0)
    return img


def to_y_channel(img: npt.NDArray) -> npt.NDArray:
    """
    Return the Y (luma) channel of a YCbCr image.

    Args:
        img (np.ndarray): input image in YCbCr format. Range must be [0, 255]

    Returns:
        np.ndarray: the luma channel of the input image.
    """
    img = img.astype(np.float32) / 255.0
    if img.ndim == 3 and img.shape[2] == 3:
        img = bgr2ycbcr(img, only_y=True)
        img = img[..., None]
    return img * 255.0


def bgr2ycbcr(img: npt.NDArray, only_y: bool = False) -> npt.NDArray:
    """
    Convert the given numpy image array from BGR to YCBCR.

    If only the Y channel is required, set only_y to true.

    Args:
        img (np.ndarray): the BGR image to convert.
        only_y (bool): if true, only the luma (Y) channel will be returned.

    Returns:
        np.ndarray: The converted image.
    """
    intype = img.dtype
    img.astype(np.float32)
    if intype != np.uint8:
        img *= 255

    if only_y:
        rlt = np.dot(img, [24.966, 128.553, 65.481]) / 255.0 + 16.0
    else:
        rlt = np.matmul(
            img,
            [
                [24.966, 112.0, -18.214],
                [128.553, -74.203, -93.786],
                [65.481, -37.797, 112.0],
            ],
        ) / 255.0 + [16, 128, 128]

    if intype == np.uint8:
        rlt = rlt.round()
    else:
        rlt /= 255.0
    return rlt.astype(intype)


def rgb2ycbcr_torch(img: torch.Tensor, only_y: bool = False) -> torch.Tensor:
    """
    Convert the given torch Tensor image array from RGB to YCBCR.

    If only the Y channel is required, set only_y to true.

    Args:
        img (torch.Tensor): the BGR image to convert.
        only_y (bool): if true, only the luma (Y) channel will be returned.

    Returns:
        torch.Tensor: The converted image.
    """
    if only_y:
        weight = torch.tensor([[65.481], [128.553], [24.966]]).to(img)
        out_img = torch.matmul(img.permute(0, 2, 3, 1), weight).permute(0, 3, 1, 2) + 16.0
    else:
        weight = torch.tensor(
            [
                [65.481, -37.797, 112.0],
                [128.553, -74.203, -93.786],
                [24.966, 112.0, -18.214],
            ]
        ).to(img)
        bias = torch.tensor([16, 128, 128]).view(1, 3, 1, 1).to(img)
        out_img = torch.matmul(img.permute(0, 2, 3, 1), weight).permute(0, 3, 1, 2) + bias

    out_img = out_img / 255.0
    return out_img



from __future__ import annotations

import abc
import copy
import dataclasses
import enum
import typing

import torch
import torch.utils.data as tud

from helios import core
from helios.core import rng

if typing.TYPE_CHECKING:
    from ..trainer import Trainer

from .samplers import (
    ResumableDistributedSampler,
    ResumableRandomSampler,
    ResumableSamplerType,
    ResumableSequentialSampler,
)

DATASET_REGISTRY = core.Registry("dataset")


def create_dataset(type_name: str, *args: typing.Any, **kwargs: typing.Any):
    """
    Create a dataset of the given type.

    This uses DATASET_REGISTRY to look-up transform types, so ensure your datasets
    have been registered before using this function.

    Args:
        type_name (str): the type of the dataset to create.
        args: positional arguments to pass into the dataset.
        kwargs: keyword arguments to pass into the dataset.

    Returns:
        nn.Module: the constructed transform.
    """
    return DATASET_REGISTRY.get(type_name)(*args, **kwargs)


class DatasetSplit(enum.Enum):
    """The different dataset splits."""

    TRAIN = 0
    VALID = 1
    TEST = 2

    @staticmethod
    def from_str(label: str) -> DatasetSplit:
        """
        Convert the given string to the corresponding enum value.

        Must be one of "train", "test", or "valid"

        Args:
            label (str): the label to convert.

        Returns:
            DatasetSplit: the corresponding value.
        """
        if label == "train":
            return DatasetSplit.TRAIN
        if label == "valid":
            return DatasetSplit.VALID
        if label == "test":
            return DatasetSplit.TEST

        raise ValueError(
            "invalid dataset split. Expected one of 'train', 'test', or "
            f"'valid', but received '{label}'"
        )


def _seed_worker(worker_id: int) -> None:
    worker_seed = torch.initial_seed() % 2**32
    rng.seed_rngs(worker_seed, skip_torch=True)


def create_dataloader(
    dataset: tud.Dataset,
    random_seed: int = 0,
    batch_size: int = 1,
    shuffle: bool = False,
    num_workers: int = 0,
    pin_memory: bool = True,
    drop_last: bool = False,
    debug_mode: bool = False,
    is_distributed: bool = False,
    sampler: ResumableSamplerType | None = None,
) -> tuple[tud.DataLoader, ResumableSamplerType]:
    """
    Create the dataloader for the given dataset.

    If no sampler is provided, the choice of sampler will be determined based on the
    values of is_distributed and shuffle. Specifically, the following logic is used:
        * If is_distributed, then sampler is ResumableDistributedSampler.
        * Otherwise, if shuffle then sampler is ResumableRandomSampler, else
        ResumableSequentialSampler.

    You may override this behaviour by providing your own sampler instance. Note that the
    sampler MUST be derived from one of ResumableSampler or ResumableDistributedSampler.

    Args:
        dataset (Dataset): the dataset.
        random_seed (int): value to use as seed for the worker processes.
        batch_size (int): number of samplers per batch.
        shuffle (bool): if true, samples are randomly shuffled.
        num_workers (int): number of worker processes for loading data.
        pin_memory (bool): if true, use page-locked device memory.
        drop_last (bool): if true, remove the final batch.
        debug_mode (bool): if true, set number of workers to 0.
        is_distributed (bool): if true, create the distributed sampler.
        sampler (ResumableSamplerType | None): (optional) sampler to use.

    Returns:
        tuple[Dataloader, Sampler]: the dataloader and sampler.
    """
    assert len(typing.cast(typing.Sized, dataset))

    if sampler is None:
        if is_distributed:
            sampler = ResumableDistributedSampler(
                dataset, shuffle=shuffle, drop_last=drop_last, seed=random_seed
            )
        else:
            if shuffle:
                sampler = ResumableRandomSampler(
                    dataset,  # type: ignore[arg-type]
                    seed=random_seed,
                    batch_size=batch_size,
                )
            else:
                sampler = ResumableSequentialSampler(
                    dataset,  # type: ignore[arg-type]
                    batch_size=batch_size,
                )

    elif not isinstance(sampler, typing.get_args(ResumableSamplerType)):
        raise TypeError(
            "error: expected sampler to derive from one of ResumableSampler or "
            f"ResumableDistributedSampler, but received {type(sampler)}"
        )

    return (
        tud.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_workers if not debug_mode else 0,
            pin_memory=pin_memory,
            drop_last=drop_last,
            sampler=sampler,
            worker_init_fn=_seed_worker,
        ),
        sampler,
    )


@dataclasses.dataclass
class DataLoaderParams:
    """
    Params used to create the dataloader object.

    Args:
        random_seed (int): value to use as seed for the worker processes.
        batch_size (int): number of samplers per batch.
        shuffle (bool): if true, samples are randomly shuffled.
        num_workers (int): number of worker processes for loading data.
        pin_memory (bool): if true, use page-locked device memory.
        drop_last (bool): if true, remove the final batch.
        debug_mode (bool): if true, set number of workers to 0.
        is_distributed (bool): if true, create the distributed sampler.
        sampler (ResumableSamplerType | None): (optional) sampler to use.
    """

    random_seed: int = rng.get_default_seed()
    batch_size: int = 1
    shuffle: bool = False
    num_workers: int = 0
    pin_memory: bool = True
    drop_last: bool = False
    debug_mode: bool = False
    is_distributed: bool | None = None
    sampler: ResumableSamplerType | None = None

    def to_dict(self) -> dict[str, typing.Any]:
        """Convert the params object to a dictionary using shallow copies."""
        return {
            field.name: getattr(self, field.name) for field in dataclasses.fields(self)
        }

    @classmethod
    def from_dict(cls, table: dict[str, typing.Any]):
        """Create a new params object from the given table."""
        params = cls()
        keys = [field.name for field in dataclasses.fields(params)]

        # Skip the sampler key, since that one needs to be created differently.
        for key in keys:
            if key in table and key != "sampler":
                setattr(params, key, table[key])

        return params


@dataclasses.dataclass
class Dataset:
    """
    The dataset and corresponding data loader params.

    Args:
        dataset (torch.utils.data.Dataset): the dataset.
        params (DataLoaderParams): the data loader params.
    """

    dataset: tud.Dataset
    params: DataLoaderParams

    def dict(self) -> dict[str, typing.Any]:
        """Convert to a dictionary."""
        res = self.params.to_dict()
        res["dataset"] = self.dataset
        return res


class DataModule(abc.ABC):
    """
    Base class that groups together the creation of the main training datasets.

    The use of this class is to standardize the way datasets and their respective
    dataloaders are created, thereby allowing consistent settings across models.
    Ex:
        ```py
        from torchvision.datasets import CIFAR10
        from helios import data

        class MyDataModule(data.DataModule):
            def prepare_data(self) -> None:
                # Use this function to prepare the data for your datasets. This will be
                # called before the distributed processes are created (if using) so you
                # should not set any state here.
                CIFAR10(download=True) # download the dataset only.

            def setup(self) -> None:
                # Create the training dataset using a DataLoaderParams instance. Note that
                # you MUST assign it to self._train_dataset.
                self._train_dataset = self._create_dataset(CIFAR10(train=True),
                                                           DataLoaderParams(...))

                # It is also possible to create a dataset using a table of key-value pairs
                # that was loaded from a config file or manually created. Let's use one to
                # create the validation split:
                settings = {"batch_size": 1, ...}

                # We can now use it to assign to self._valid_dataset like this:
                self._valid_dataset = self._create_dataset(CIFAR10(train=False), settings)

                # Finally, if you need a testing split, you can create it like this:
                self._test_dataset = self._create_dataset(CIFAR10(train=False), settings)

            def teardown(self) -> None:
                # Use this function to clean up any state. It will be called after
                # training is done.
        ```
    """

    def __init__(self) -> None:
        """Create the data module."""
        self._is_distributed: bool = False
        self._train_dataset: Dataset | None = None
        self._valid_dataset: Dataset | None = None
        self._test_dataset: Dataset | None = None
        self._trainer: Trainer | None = None

    @property
    def is_distributed(self) -> bool:
        """Flag controlling whether distributed training is being used or not."""
        return self._is_distributed

    @is_distributed.setter
    def is_distributed(self, val: bool) -> None:
        self._is_distributed = val

    @property
    def trainer(self) -> Trainer:
        """Reference to the trainer."""
        return core.get_from_optional(self._trainer)

    @trainer.setter
    def trainer(self, t) -> None:
        self._trainer = t

    def prepare_data(self) -> None:  # noqa: B027
        """
        Prepare data for training.

        This can include downloading datasets, or streaming them from external services.
        This function will be called on the primary process when using distributed
        training (will be called prior to initialization of the processes) so don't store
        any state here.
        """

    @abc.abstractmethod
    def setup(self) -> None:
        """Construct all required datasets."""

    def train_dataloader(self) -> tuple[tud.DataLoader, ResumableSamplerType] | None:
        """Create the train dataloader (if available)."""
        if self._train_dataset is None:
            return None
        return self._create_dataloader(self._train_dataset)

    def valid_dataloader(self) -> tuple[tud.DataLoader, ResumableSamplerType] | None:
        """Create the valid dataloader (if available)."""
        if self._valid_dataset is None:
            return None
        return self._create_dataloader(self._valid_dataset)

    def test_dataloader(self) -> tuple[tud.DataLoader, ResumableSamplerType] | None:
        """Create the test dataloader (if available)."""
        if self._test_dataset is None:
            return None
        return self._create_dataloader(self._test_dataset)

    def teardown(self) -> None:  # noqa: B027
        """Clean up any state after training is over."""

    def _create_dataset(
        self, dataset: tud.Dataset, params: DataLoaderParams | dict[str, typing.Any]
    ) -> Dataset:
        """
        Create a dataset object from a Torch dataset and a params.

        This is a convenience function to help you create the dataset objects. The params
        object can either be the fully filled in DataLoaderParams instance, or it can be a
        dict containing all of the necessary settings. If a dict is passed in, it will be
        used to populate the corresponding DataLoaderParams object.

        Args:
            dataset (torch.utils.data.Dataset): the dataset.
            params (DataLoaderParams | dict[str, Any]): either a params object or a dict.

        Returns:
            Dataset: the new dataset object.
        """
        return Dataset(
            dataset,
            copy.deepcopy(params)
            if isinstance(params, DataLoaderParams)
            else DataLoaderParams.from_dict(params),
        )

    def _create_dataloader(
        self, dataset: Dataset
    ) -> tuple[tud.DataLoader, ResumableSamplerType]:
        # Only override the distributed flag if it hasn't been set by the user.
        if dataset.params.is_distributed is None:
            dataset.params.is_distributed = self._is_distributed
        return create_dataloader(**dataset.dict())



import typing

from torch import nn, optim

from helios import core


def _register_default_optimizers(registry: core.Registry) -> None:
    """
    Register the default optimizers.

    Args:
        registry (Registry): the optimizer registry.
    """
    registry.register(optim.Adam)
    registry.register(optim.AdamW)
    registry.register(optim.SGD)


OPTIMIZER_REGISTRY = core.Registry("optimizer")
_register_default_optimizers(OPTIMIZER_REGISTRY)


def create_optimizer(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> nn.Module:
    """
    Create the optimizer for the given type.

    Args:
        type_name (str): the type of the optimizer to create.
        args: positional arguments to pass into the optimizer.
        kwargs: keyword arguments to pass into the optimizer.
    Returns:
        nn.Module: the optimizer.
    """
    return OPTIMIZER_REGISTRY.get(type_name)(*args, **kwargs)



from __future__ import annotations

import abc
import typing

import torch

from helios import core
from helios.core import distributed as dist

if typing.TYPE_CHECKING:
    from ..trainer import Trainer, TrainingState


# Tell Ruff to ignore the empty-method-without-abstract-decorator check, since a lot of
# functions in the Model base class will be empty by default.
# ruff: noqa: B027


class Model(abc.ABC):
    """Base class for training models."""

    def __init__(self, save_name: str):
        """Create the model."""
        self._save_name: str = save_name
        self._trainer: Trainer | None

        self._is_distributed: bool = False
        self._map_loc: str | dict[str, str] = ""
        self._device: torch.device | None = None
        self._rank: int = 0

        self._loss_items: dict[str, torch.Tensor] = {}
        self._running_losses: dict[str, float] = {}
        self._val_scores: dict[str, float] = {}
        self._test_scores: dict[str, float] = {}

    @property
    def save_name(self) -> str:
        """The name of the model used for saving checkpoints and final networks."""
        return self._save_name

    @property
    def is_distributed(self) -> bool:
        """Flag controlling whether distributed training is being used or not."""
        return self._is_distributed

    @is_distributed.setter
    def is_distributed(self, val: bool) -> None:
        self._is_distributed = val

    @property
    def map_loc(self) -> str | dict[str, str]:
        """The location to map loaded weights from a checkpoint or pre-trained file."""
        return self._map_loc

    @map_loc.setter
    def map_loc(self, loc: str | dict[str, str]) -> None:
        self._map_loc = loc

    @property
    def device(self) -> torch.device:
        """The device on which the tensors of the model are mapped to."""
        return core.get_from_optional(self._device)

    @device.setter
    def device(self, dev: torch.device) -> None:
        self._device = dev

    @property
    def rank(self) -> int:
        """The local rank (device id) that the model is running on."""
        return self._rank

    @rank.setter
    def rank(self, r: int) -> None:
        self._rank = r

    @property
    def trainer(self) -> Trainer:
        """Reference to the trainer."""
        return core.get_from_optional(self._trainer)

    @trainer.setter
    def trainer(self, t) -> None:
        self._trainer = t

    @abc.abstractmethod
    def setup(self, fast_init: bool = False) -> None:
        """
        Initialize all the state necessary for training.

        Use this function to load all the networks, schedulers, optimizers, losses, etc.
        that you require for training. This will be called before training starts and
        after the distributed processes have been launched (if applicable).

        The fast_init flag is used to indicate that the model is going to be used to strip
        training data from any saved checkpoints or for testing. As such, you should
        ONLY load the network(s) and nothing else.

        Args:
            fast_init (bool): if True, only networks are loaded.
        """

    def load_for_testing(self) -> None:
        """
        Load the network(s) used for testing.

        When testing, the trainer will try to load the last checkpoint saved during
        training. If it is unable to find one, it will call this function to ensure the
        model loads the state of the network(s) for testing. If you wish to load your
        network using separate logic, use this function.
        """

    def load_state_dict(
        self, state_dict: dict[str, typing.Any], fast_init: bool = False
    ) -> None:
        """
        Load the model state from the given state dictionary.

        Use this function to restore any training state from a checkpoint. Note that any
        weights will have been automatically mapped to the correct device.

        The fast_init flag is used to indicate that the model is going to be used to strip
        training data from any saved checkpoints or for testing. As such, you should
        ONLY load the network(s) and nothing else.

        Args:
            state_dict (dict[str, Any]): the state dictionary to load from.
            fast_init (bool): if True, only networks need to be loaded.
        """

    def state_dict(self) -> dict[str, typing.Any]:
        """
        Get the state dictionary of the model.

        Use this function to save any state that you require for checkpoints.

        Returns:
            dict[str, Any]: the state dictionary of the model.
        """
        return {}

    def trained_state_dict(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> dict[str, typing.Any]:
        """
        Get the state dictionary for the trained model.

        Use this function to save the state required for the final trained model. The
        returned dictionary should contain only the necessary information to re-create the
        network(s) along with any additional data you require.

        Args:
            args: positional arguments.
            kwargs: keyword arguments.

        Returns:
            dict[str, Any]: the state dictionary without any training data.
        """
        return {}

    def append_metadata_to_chkpt_name(self, chkpt_name: str) -> str:
        """
        Append additional data to the checkpoint filename.

        Use this function to append the value of the loss function(s), validation
        metric(s), or any extra metadata you wish to add to the name of the checkpoint.
        Note that the epoch and iteration numbers are added automatically. The extension
        will also be added automatically.

        Args:
            chkpt_name (str): the name of the checkpoint filename (without extension).

        Returns:
            str: the name with any additional metadata.
        """
        return chkpt_name

    def train(self) -> None:
        """Switch the model to training mode."""

    def on_training_start(self) -> None:
        """
        Perform any necessary actions when training starts.

        You may use this function to log the network architecture, hyper-params, etc.
        """

    def on_training_batch_start(self, state: TrainingState) -> None:
        """
        Perform any actions when a training batch is started.

        This function is called before train_step is called. By default, it will clear out
        the loss table, but you may also use it to do any additional tasks prior to the
        training step itself.

        Args:
            state (TrainingState): the current training state.
        """
        self._loss_items.clear()

    @abc.abstractmethod
    def train_step(self, batch: typing.Any, state: TrainingState) -> None:
        """
        Perform a single training step.

        The input is the returned value from the datasets you supplied to the trainer. In
        this function, you should perform the forward and backward passes for your
        network(s). If you use schedulers, they should be updated here as well. Note that
        you do not have to clear the losses or gather them. This will be handled
        automatically for you. Also keep in mind that the contents of the batch have NOT
        been moved to the current device. It is your responsibility to do so.

        Args:
            batch (Any): the batch data returned from the dataset.
            state (TrainingState): the current training state.
        """

    def on_training_batch_end(
        self, state: TrainingState, should_log: bool = False
    ) -> None:
        """
        Perform any actions when a training batch ends.

        This function is called after train_step is called. By default, it will gather all
        the losses stored in self._loss_items (if using distributed training) and will
        update the running losses using those values. You may also use this function to
        log your losses or perform any additional tasks after the training step.

        Args:
            state (TrainingState): the current training state.
            should_log (bool): if true, then logging should be performed.
        """
        if self._is_distributed:
            for _, loss in self._loss_items.items():
                dist.all_reduce_tensors(loss)

        for key, loss in self._loss_items.items():
            if loss is not None:
                if key not in self._running_losses:
                    self._running_losses[key] = loss.item()
                else:
                    self._running_losses[key] += loss.item()

    def on_training_end(self) -> None:
        """
        Perform any necessary actions when training ends.

        You may use this function to update any weight averaging networks, or any other
        tasks that should only happen at the end of training.
        """

    def eval(self) -> None:
        """Switch the model to evaluation mode."""

    def on_validation_start(self, validation_cycle: int) -> None:
        """
        Perform any necessary actions when validation starts.

        By default, this will clear out the table of validation values, but you may use it
        for any other tasks that should happen when validation begins.

        Args:
            validation_cycle (int): the validation cycle number.
        """
        self._val_scores.clear()

    def on_validation_batch_start(self, step: int) -> None:
        """
        Perform any actions when a validation batch is started.

        This function is called before valid_step is called. No steps are performed by
        default.

        Args:
            step (int): the current validation batch.
        """

    def valid_step(self, batch: typing.Any, step: int) -> None:
        """
        Perform a single validation step.

        The input is the returned value from the datasets you supplied to the trainer. In
        this function, you should perform any steps necessary to compute the validation
        metric(s) for your network(s).

        Args:
            batch (Any): the batch data returned from the dataset.
            step (int): the current validation batch.
        """

    def on_validation_batch_end(self, step: int) -> None:
        """
        Perform any actions when a validation batch ends.

        This function is called after valid_step is called. No steps are performed by
        default.

        Args:
            step (int): the current validation batch.
        """

    def on_validation_end(self, validation_cycle: int) -> None:
        """
        Perform any necessary actions when validation ends.

        By default, this function will clear out the running loss table, but you may use
        this function to compute any final validation metrics as well as log them.

        Args:
            validation_cycle (int): the validation cycle number.
        """
        self._running_losses.clear()

    def have_metrics_improved(self) -> bool:
        """
        Determine whether the current validation results are an improvement or not.

        This is used when early stopping is enabled in the trainer to determine whether
        the stop cycle count should increase or not.

        Returns:
            bool: false if no improvements were seen in the last validation cycle.
        """
        return True

    def on_testing_start(self) -> None:
        """
        Perform any necessary actions when testing starts.

        By default, this will clear out the table of testing values, but you may use it
        for any other tasks that should happen when testing begins.
        """
        self._test_scores.clear()

    def on_testing_batch_start(self, step: int) -> None:
        """
        Perform any actions when a testing batch is started.

        This function is called before test_step is called. No steps are performed by
        default.

        Args:
            step (int): the current testing batch.
        """

    def test_step(self, batch: typing.Any, step: int) -> None:
        """
        Perform a single testing step.

        The input is the returned value from the datasets you supplied to the trainer. In
        this function, you should perform any steps necessary to compute the testing
        metric(s) for your network(s).

        Args:
            batch (Any): the batch data returned from the dataset.
            step (int): the current validation batch.
        """

    def on_testing_batch_end(self, step: int) -> None:
        """
        Perform any actions when a testing batch ends.

        This function is called after test_step is called. No steps are performed by
        default.

        Args:
            step (int): the current testing batch.
        """

    def on_testing_end(self) -> None:
        """
        Perform any necessary actions when testing ends.

        You may use this function to compute any final testing metrics as well as log
        them.
        """



import typing

import torch
from torch.utils import data as tud

from helios import core, data
from helios.core import rng
from helios.data import samplers as hlds
from helios.data import transforms as hldt

DATASET_SIZE: int = 16


class RandomDataset(tud.Dataset):
    def __getitem__(self, index):
        gen = rng.get_default_numpy_rng().generator
        return gen.integers(0, 1000, 3)

    def __len__(self):
        return DATASET_SIZE


class SequentialDataset(tud.Dataset):
    def __init__(self):
        super().__init__()

        self._samples = [
            torch.tensor([359, 639, 636]),
            torch.tensor([963, 738, 996]),
            torch.tensor([766, 97, 292]),
            torch.tensor([500, 324, 687]),
            torch.tensor([15, 841, 180]),
            torch.tensor([553, 266, 928]),
            torch.tensor([269, 806, 472]),
            torch.tensor([694, 632, 678]),
            torch.tensor([92, 228, 424]),
            torch.tensor([421, 568, 705]),
            torch.tensor([686, 769, 903]),
            torch.tensor([440, 610, 126]),
            torch.tensor([809, 833, 629]),
            torch.tensor([340, 810, 512]),
            torch.tensor([23, 322, 577]),
            torch.tensor([757, 651, 429]),
        ]

    def __getitem__(self, index):
        return self._samples[index]

    def __len__(self):
        return DATASET_SIZE


class SampleDataModule(data.DataModule):
    def setup(self) -> None:
        params = data.DataLoaderParams(
            batch_size=2, num_workers=2, random_seed=rng.get_default_seed()
        )

        params.shuffle = True
        self._train_dataset = self._create_dataset(RandomDataset(), params)

        params.num_workers = 2
        self._valid_dataset = self._create_dataset(SequentialDataset(), params)

        params.shuffle = False
        self._test_dataset = self._create_dataset(SequentialDataset(), params)


class TestTransforms:
    def test_to_tensor(self) -> None:
        assert "ToTensor" in hldt.TRANSFORM_REGISTRY

        rng.seed_rngs()
        gen = rng.get_default_numpy_rng().generator
        img = gen.random(size=(32, 32, 3))

        to_tensor = hldt.create_transform("ToTensor")
        as_tens = to_tensor(img)

        assert isinstance(as_tens, torch.Tensor)
        assert as_tens.shape == (3, 32, 32)


class TestDataModule:
    def get_expected_train_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[359, 639, 636], [963, 738, 996]]),
            torch.tensor([[766, 97, 292], [500, 324, 687]]),
            torch.tensor([[15, 841, 180], [553, 266, 928]]),
            torch.tensor([[269, 806, 472], [694, 632, 678]]),
            torch.tensor([[92, 228, 424], [421, 568, 705]]),
            torch.tensor([[686, 769, 903], [440, 610, 126]]),
            torch.tensor([[809, 833, 629], [340, 810, 512]]),
            torch.tensor([[23, 322, 577], [757, 651, 429]]),
        ]

    def get_expected_valid_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[809, 833, 629], [694, 632, 678]]),
            torch.tensor([[359, 639, 636], [553, 266, 928]]),
            torch.tensor([[500, 324, 687], [963, 738, 996]]),
            torch.tensor([[421, 568, 705], [15, 841, 180]]),
            torch.tensor([[766, 97, 292], [23, 322, 577]]),
            torch.tensor([[269, 806, 472], [757, 651, 429]]),
            torch.tensor([[340, 810, 512], [440, 610, 126]]),
            torch.tensor([[92, 228, 424], [686, 769, 903]]),
        ]

    def get_expected_test_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[359, 639, 636], [963, 738, 996]]),
            torch.tensor([[766, 97, 292], [500, 324, 687]]),
            torch.tensor([[15, 841, 180], [553, 266, 928]]),
            torch.tensor([[269, 806, 472], [694, 632, 678]]),
            torch.tensor([[92, 228, 424], [421, 568, 705]]),
            torch.tensor([[686, 769, 903], [440, 610, 126]]),
            torch.tensor([[809, 833, 629], [340, 810, 512]]),
            torch.tensor([[23, 322, 577], [757, 651, 429]]),
        ]

    def check_batches(self, exp: list[torch.Tensor], val: list[torch.Tensor]) -> None:
        for b, exp_b in zip(exp, val, strict=True):
            assert torch.all(b == exp_b)

    def prepare(
        self, split: data.DatasetSplit, skip_seed: bool = False
    ) -> tuple[tud.DataLoader, tud.Sampler]:
        if not skip_seed:
            rng.seed_rngs()

        datamodule = SampleDataModule()
        datamodule.setup()

        if split == data.DatasetSplit.TRAIN:
            return core.get_from_optional(datamodule.train_dataloader())
        if split == data.DatasetSplit.VALID:
            return core.get_from_optional(datamodule.valid_dataloader())
        return core.get_from_optional(datamodule.test_dataloader())

    def test_worker_seeding(self) -> None:
        dataloader, _ = self.prepare(data.DatasetSplit.TRAIN)

        batches = []
        for batch in dataloader:
            batches.append(batch)

        exp_batches = self.get_expected_train_batches()
        self.check_batches(exp_batches, batches)

    def test_resume_random(self) -> None:
        dataloader, sampler = self.prepare(data.DatasetSplit.VALID)
        assert isinstance(sampler, hlds.ResumableRandomSampler)

        half_step = len(dataloader) // 2
        exp_batches = self.get_expected_valid_batches()

        batches = []
        sampler.set_epoch(0)
        for step, batch in enumerate(dataloader):
            if step >= half_step:
                break
            batches.append(batch)

        self.check_batches(exp_batches[:half_step], batches)

        # Grab the state and clear everything.
        rng_state = rng.get_rng_state_dict()

        del dataloader, sampler
        batches = []
        rng.seed_rngs(0)

        # Restore everything.
        rng.load_rng_state_dict(rng_state)
        dataloader, sampler = self.prepare(data.DatasetSplit.VALID, skip_seed=True)
        assert isinstance(sampler, hlds.ResumableRandomSampler)
        sampler.set_epoch(0)
        sampler.start_iter = half_step

        # Continue.
        for batch in dataloader:
            batches.append(batch)

        self.check_batches(exp_batches[half_step:], batches)

    def test_resume_sequential(self) -> None:
        dataloader, sampler = self.prepare(data.DatasetSplit.TEST)
        assert isinstance(sampler, hlds.ResumableSequentialSampler)

        half_step = len(dataloader) // 2
        exp_batches = self.get_expected_test_batches()

        batches = []
        sampler.set_epoch(0)
        for step, batch in enumerate(dataloader):
            if step >= half_step:
                break
            batches.append(batch)

        self.check_batches(exp_batches[:half_step], batches)

        # Grab the state and clear everything.
        rng_state = rng.get_rng_state_dict()

        del dataloader, sampler
        batches = []
        rng.seed_rngs(0)

        # Restore everything.
        rng.load_rng_state_dict(rng_state)
        dataloader, sampler = self.prepare(data.DatasetSplit.TEST, skip_seed=True)
        assert isinstance(sampler, hlds.ResumableSequentialSampler)
        sampler.set_epoch(0)
        sampler.start_iter = half_step

        # Continue.
        for batch in dataloader:
            batches.append(batch)

        self.check_batches(exp_batches[half_step:], batches)

    def test_sampler_registry(self) -> None:
        registered_names = [
            "ResumableRandomSampler",
            "ResumableSequentialSampler",
            "ResumableDistributedSampler",
        ]
        assert len(typing.cast(typing.Sized, hlds.SAMPLER_REGISTRY.keys())) == len(
            registered_names
        )
        for name in registered_names:
            assert name in hlds.SAMPLER_REGISTRY


if __name__ == "__main__":
    t = TestDataModule()
    t.test_resume_sequential()



import cv2
import numpy as np
import numpy.typing as npt
import torch
import torch.nn.functional as F

from helios.data import functional


def _ssim(img: npt.NDArray[np.float64], img2: npt.NDArray[np.float64]) -> float:
    """
    SSIM implementation function. See calculate_ssim for details.

    Args:
        img (np.ndarray): Images with range [0, 1] (float64).
        img2 (np.ndarray): Images with range [0, 1] (float64).

    Returns:
        float: SSIM.
    """
    c1 = (0.01 * 255) ** 2
    c2 = (0.03 * 255) ** 2
    kernel: npt.NDArray = cv2.getGaussianKernel(11, 1.5)
    window = np.outer(kernel, kernel.transpose())

    # valid mode for window size 11
    mu1: npt.NDArray = cv2.filter2D(img, -1, window)[5:-5, 5:-5]
    mu2: npt.NDArray = cv2.filter2D(img2, -1, window)[5:-5, 5:-5]
    mu1_sq = mu1**2
    mu2_sq = mu2**2
    mu1_mu2 = mu1 * mu2
    sigma1_sq = cv2.filter2D(img**2, -1, window)[5:-5, 5:-5] - mu1_sq  # type: ignore[operator]
    sigma2_sq = cv2.filter2D(img2**2, -1, window)[5:-5, 5:-5] - mu2_sq  # type: ignore[operator]
    sigma12 = cv2.filter2D(img * img2, -1, window)[5:-5, 5:-5] - mu1_mu2

    ssim_map = ((2 * mu1_mu2 + c1) * (2 * sigma12 + c2)) / (
        (mu1_sq + mu2_sq + c1) * (sigma1_sq + sigma2_sq + c2)
    )
    return ssim_map.mean()


def _ssim_torch(img: torch.Tensor, img2: torch.Tensor) -> float:
    """
    SSIM Torch implementation function. See calculate_ssim_torch for details.

    Args:
        img (torch.Tensor): Images with range [0, 1] (float64).
        img2 (torch.Tensor): Images with range [0, 1] (float64).

    Returns:
        float: SSIM.
    """
    c1 = (0.01 * 255) ** 2
    c2 = (0.03 * 255) ** 2

    kernel: npt.NDArray = cv2.getGaussianKernel(11, 1.5)
    window = (
        torch.from_numpy(np.outer(kernel, kernel.transpose()))
        .view(1, 1, 11, 11)
        .expand(img.size(1), 1, 11, 11)
        .to(img.dtype)
        .to(img.device)
    )

    # pylint: disable=not-callable
    mu1 = F.conv2d(img, window, stride=1, padding=0, groups=img.shape[1])  # valid mode
    mu2 = F.conv2d(img2, window, stride=1, padding=0, groups=img2.shape[1])  # valid mode
    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2
    sigma1_sq = (
        F.conv2d(img * img, window, stride=1, padding=0, groups=img.shape[1]) - mu1_sq
    )
    sigma2_sq = (
        F.conv2d(img2 * img2, window, stride=1, padding=0, groups=img.shape[1]) - mu2_sq
    )
    sigma12 = (
        F.conv2d(img * img2, window, stride=1, padding=0, groups=img.shape[1]) - mu1_mu2
    )

    cs_map = (2 * sigma12 + c2) / (sigma1_sq + sigma2_sq + c2)
    ssim_map = ((2 * mu1_mu2 + c1) / (mu1_sq + mu2_sq + c1)) * cs_map
    return float(ssim_map.mean([1, 2, 3]))


def _average_precision(
    output: npt.NDArray[np.float64], target: npt.NDArray[np.float64]
) -> float:
    """
    Calculate the average precision of the given inputs. See calculate_map for details.

    Args:
        output (np.ndarray): predicate labels in range [0, 1].
        target (np.ndarray): target (inferred) labels in range [0, 1].

    Returns:
        float: the average precision.
    """
    epsilon = 1e-8

    indices = output.argsort()[::-1]
    total_count = np.cumsum(np.ones((len(output), 1)))

    target_ = target[indices]
    ind = target_ == 1
    pos_count = np.cumsum(ind)
    total = pos_count[-1]
    pos_count[np.logical_not(ind)] = 0
    pp = pos_count / total_count
    precision_at_i = np.sum(pp)
    precision_at_i = precision_at_i / (total + epsilon)

    return precision_at_i


def _mae_torch(pred: torch.Tensor, gt: torch.Tensor) -> float:
    """
    MAE Torch implementation. See calculate_mae_torch for details.

    Args:
        pred (torch.Tensor): predicate (inferred) images in range [0, 255]
        gt (torch.Tensor): ground-truth images in range [0, 255]

    Returns:
        float: the MAE score.
    """
    h, w = gt.shape[0:2]
    sum_error = torch.sum(torch.absolute(torch.sub(pred.float(), gt.float())))
    mae_error = torch.divide(sum_error, float(h) * float(w) * 2550 + 1e-4)
    return float(mae_error)


def _f1_score_torch(
    pd: torch.Tensor, gt: torch.Tensor
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Calculate the following scores: precision, recall, and F1 for torch.

    See calculate_precision/recall/f1_torch for details.

    Args:
        pd (torch.Tensor): predicate tensor.
        gt (torch.Tensor): ground-truth tensor.

    Returns:
        tuple[torch.Tensor, torch.Tensor, torch.Tensor]: the precision, recall, and F1
        scores, respectively.
    """
    gt_num = torch.sum((gt > 128).float() * 1)

    pp = pd[gt > 128]
    nn = pd[gt <= 128]

    pp_hist = torch.histc(pp, bins=255, min=0, max=255)
    nn_hist = torch.histc(nn, bins=255, min=0, max=255)

    pp_hist_flip = torch.flipud(pp_hist)
    nn_hist_flip = torch.flipud(nn_hist)

    pp_hist_flip_cum = torch.cumsum(pp_hist_flip, dim=0)
    nn_hist_flip_cum = torch.cumsum(nn_hist_flip, dim=0)

    precision = (pp_hist_flip_cum) / (pp_hist_flip_cum + nn_hist_flip_cum + 1e-4)
    recall = (pp_hist_flip_cum) / (gt_num + 1e-4)
    f1 = (1 + 0.3) * precision * recall / (0.3 * precision + recall + 1e-4)

    return (
        torch.reshape(precision, (1, precision.shape[0])),
        torch.reshape(recall, (1, recall.shape[0])),
        torch.reshape(f1, (1, f1.shape[0])),
    )


def calculate_psnr(
    img: npt.NDArray,
    img2: npt.NDArray,
    crop_border: int,
    input_order: str = "HWC",
    test_y_channel: bool = False,
) -> float:
    """
    Calculate PSNR (Peak Signal-to-Noise Ratio).

    Implementation follows: https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio

    Args:
        img (np.ndarray): Images with range [0, 255].
        img2 (np.ndarray): Images with range [0, 255].
        crop_border (int): Cropped pixels in each edge of an image. These pixels are not
        involved in the calculation.
        input_order (str): Whether the input order is 'HWC' or 'CHW'.
        test_y_channel (bool): Test on Y channel of YCbCr. Default:

    Returns:
        float: PSNR value.
    """
    assert (
        img.shape == img2.shape
    ), f"Image shapes are different: {img.shape}, {img2.shape}."

    img = functional.convert_to_hwc(img, input_order=input_order)
    img2 = functional.convert_to_hwc(img2, input_order=input_order)

    if crop_border != 0:
        img = img[crop_border:-crop_border, crop_border:-crop_border, ...]
        img2 = img2[crop_border:-crop_border, crop_border:-crop_border, ...]

    if test_y_channel:
        img = functional.to_y_channel(img)
        img2 = functional.to_y_channel(img2)

    img = img.astype(np.float64)
    img2 = img2.astype(np.float64)

    mse = np.mean((img - img2) ** 2)
    if mse == 0:
        return float("inf")
    return 10.0 * np.log10(255.0 * 255.0 / mse)


def calculate_psnr_torch(
    img: torch.Tensor,
    img2: torch.Tensor,
    crop_border: int,
    test_y_channel: bool = False,
) -> float:
    """
    Calculate PSNR (Peak Signal-to-Noise Ratio).

    Implementation follows: https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio

    Args:
        img (torch.Tensor): Images with range [0, 255].
        img2 (torch.Tensor): Images with range [0, 255].
        crop_border (int): Cropped pixels in each edge of an image. These pixels are not
        involved in the calculation.
        test_y_channel (bool): Test on Y channel of YCbCr. Default:

    Returns:
        float: PSNR value.
    """
    assert (
        img.shape == img2.shape
    ), f"Image shapes are different: {img.shape}, {img2.shape}."

    if crop_border != 0:
        img = img[:, :, crop_border:-crop_border, crop_border:-crop_border]
        img2 = img2[:, :, crop_border:-crop_border, crop_border:-crop_border]

    if test_y_channel:
        img = functional.rgb2ycbcr_torch(img, only_y=True)
        img2 = functional.rgb2ycbcr_torch(img2, only_y=True)

    img = img.to(torch.float64)
    img2 = img2.to(torch.float64)

    mse = torch.mean((img - img2) ** 2, dim=[1, 2, 3])
    return float(10.0 * torch.log10(1.0 / (mse + 1e-8)))


def calculate_ssim(
    img: npt.NDArray,
    img2: npt.NDArray,
    crop_border: int,
    input_order: str = "HWC",
    test_y_channel: bool = False,
) -> float:
    """
    Calculate SSIM (structural similarity).

    Implementation follows: 'Image quality assesment: From error visibility to structural
    similarity'. Results are identical to those of the official MATLAB code in
    https://ece.uwaterloo.ca/~z70wang/research/ssim/.
    For three-channel images, SSIM is calculated for each channel and then
    averaged.

    Args:
        img (np.ndarray): Images with range [0, 255].
        img2 (np.ndarray): Images with range [0, 255].
        crop_border (int): Cropped pixels in each edge of an image. These pixels are not
        involved in the calculation.
        input_order (str): Whether the input order is 'HWC' or 'CHW'.
        test_y_channel (bool): Test on Y channel of YCbCr.

    Returns:
        float: SSIM.
    """
    assert (
        img.shape == img2.shape
    ), f"Image shapes are different: {img.shape}, {img2.shape}."

    img = functional.convert_to_hwc(img, input_order=input_order)
    img2 = functional.convert_to_hwc(img2, input_order=input_order)

    if crop_border != 0:
        img = img[crop_border:-crop_border, crop_border:-crop_border, ...]
        img2 = img2[crop_border:-crop_border, crop_border:-crop_border, ...]

    if test_y_channel:
        img = functional.to_y_channel(img)
        img2 = functional.to_y_channel(img2)

    img = img.astype(np.float64)
    img2 = img2.astype(np.float64)

    ssims = []
    for i in range(img.shape[2]):
        ssims.append(_ssim(img[..., i], img2[..., i]))
    return np.array(ssims).mean()


def calculate_ssim_torch(
    img: torch.Tensor,
    img2: torch.Tensor,
    crop_border: int,
    test_y_channel: bool = False,
) -> float:
    """
    Calculate SSIM (structural similarity).

    Implementation follows: 'Image quality assesment: From error visibility to structural
    similarity'. Results are identical to those of the official MATLAB code in
    https://ece.uwaterloo.ca/~z70wang/research/ssim/.
    For three-channel images, SSIM is calculated for each channel and then
    averaged.

    Args:
        img (torch.Tensor): Images with range [0, 255].
        img2 (torch.Tensor): Images with range [0, 255].
        crop_border (int): Cropped pixels in each edge of an image. These pixels are not
        involved in the calculation.
        test_y_channel (bool): Test on Y channel of YCbCr.

    Returns:
        float: the SSIM value.
    """
    assert (
        img.shape == img2.shape
    ), f"Image shapes are different: {img.shape}, {img2.shape}."

    if crop_border != 0:
        img = img[:, :, crop_border:-crop_border, crop_border:-crop_border]
        img2 = img2[:, :, crop_border:-crop_border, crop_border:-crop_border]

    if test_y_channel:
        img = functional.rgb2ycbcr_torch(img, only_y=True)
        img2 = functional.rgb2ycbcr_torch(img2, only_y=True)

    img = img.to(torch.float64)
    img2 = img2.to(torch.float64)

    ssim = _ssim_torch(img * 255.0, img2 * 255.0)
    return ssim


def calculate_mAP(targs: npt.NDArray, preds: npt.NDArray) -> float:
    """
    Calculate the mAP (Mean Average Precision).

    Implementation follows:
    https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision

    Args:
        targs (np.ndarray): target (inferred) labels in range [0, 1].
        preds (np.ndarray): predicate labels in range [0, 1].

    Returns:
        float: the mAP score
    """
    if np.size(preds) == 0:
        return 0

    class_count = 0
    mAP_sum = 0.0
    for k in range(preds.shape[1]):
        scores = preds[:, k]
        targets = targs[:, k]
        if np.sum(targets) == 0:
            continue
        mAP_sum += _average_precision(scores, targets)
        class_count += 1

    return 100 * (mAP_sum / class_count)


def calculate_mae_torch(pred: torch.Tensor, gt: torch.Tensor) -> float:
    """
    Compute the MAE (Mean-Average Precision) score.

    Implementation follows: https://en.wikipedia.org/wiki/Mean_absolute_error

    Args:
        pred (torch.Tensor): predicate (inferred) images in range [0, 255]
        gt (torch.Tensor): ground-truth images in range [0, 255]

    Returns:
        float: the MAE score.
    """
    return _mae_torch(pred, gt)


def calculate_precision_torch(pred: torch.Tensor, gt: torch.Tensor) -> float:
    """
    Compute the Precision score.

    Implementation follows:  https://en.wikipedia.org/wiki/Precision_and_recall

    Args:
        pred (torch.Tensor): predicate (inferred) images in range [0, 255]
        gt (torch.Tensor): ground-truth images in range [0, 255]

    Returns:
        float: the precision score.
    """
    return _f1_score_torch(pred, gt)[0].cpu().data.numpy()


def calculate_recall_torch(pred: torch.Tensor, gt: torch.Tensor) -> float:
    """
    Compute the Recall score.

    Implementation follows:  https://en.wikipedia.org/wiki/Precision_and_recall

    Args:
        pred (torch.Tensor): predicate (inferred) images in range [0, 255]
        gt (torch.Tensor): ground-truth images in range [0, 255]

    Returns:
        float: the recall score.
    """
    return _f1_score_torch(pred, gt)[1].cpu().data.numpy()


def calculate_f1_torch(pred: torch.Tensor, gt: torch.Tensor) -> float:
    """
    Compute the F1 score.

    Implementation follows: https://en.wikipedia.org/wiki/F-score

    Args:
        pred (torch.Tensor): predicate (inferred) images in range [0, 255]
        gt (torch.Tensor): ground-truth images in range [0, 255]

    Returns:
        float: the F1 score.
    """
    return _f1_score_torch(pred, gt)[2].cpu().data.numpy()



import typing

from torch import nn

from helios import core

LOSS_REGISTRY = core.Registry("loss")


def create_loss(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> nn.Module:
    """
    Create the loss function for the given type.

    Args:
        type_name (str): the type of the loss to create.
        args: positional arguments to pass into the loss.
        kwargs: keyword arguments to pass into the loss.
    Returns:
        nn.Module: the loss function.
    """
    return LOSS_REGISTRY.get(type_name)(*args, **kwargs)



import typing

from torch import nn
from torch.optim import lr_scheduler

from helios import core


def _register_default_schedulers(registry: core.Registry):
    """
    Register the default schedulers to the registry.

    Args:
        registry (Registry): the scheduler registry.
    """
    registry.register(lr_scheduler.MultiStepLR)
    registry.register(lr_scheduler.CosineAnnealingLR)


SCHEDULER_REGISTRY = core.Registry("scheduler")
_register_default_schedulers(SCHEDULER_REGISTRY)


def create_scheduler(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> nn.Module:
    """
    Create the scheduler for the given type.

    Args:
        type_name (str): the type of the scheduler to create.
        args: positional arguments to pass into the scheduler.
        kwargs: keyword arguments to pass into the scheduler.

    Returns:
        nn.Module: the scheduler.
    """
    return SCHEDULER_REGISTRY.get(type_name)(*args, **kwargs)



import copy
import pathlib

import numpy as np
import numpy.typing as npt
import pytest
import torch
from torch.utils import data as tud

import helios.trainer as hlt
from helios import data
from helios import model as hlm
from helios.core import rng
from helios.data import functional as F

# Ignore the use of private members so we can test them correctly.
# ruff: noqa: SLF001

DATASET_SIZE = 10


class RandomDataset(tud.Dataset):
    def __getitem__(self, index):
        gen = rng.get_default_numpy_rng().generator
        return gen.integers(0, 100, 3)

    def __len__(self):
        return DATASET_SIZE


class RandomDatamodule(data.DataModule):
    def setup(self) -> None:
        params = data.DataLoaderParams(
            batch_size=1, num_workers=0, random_seed=rng.get_default_seed()
        )

        self._train_dataset = self._create_dataset(RandomDataset(), params)
        self._valid_dataset = self._create_dataset(RandomDataset(), params)
        self._test_dataset = self._create_dataset(RandomDataset(), params)


class CheckFunModel(hlm.Model):
    def __init__(self) -> None:
        super().__init__("test-model")
        self.called_train_funs: dict[str, bool] = {
            "setup": False,
            "train": False,
            "on_training_start": False,
            "on_training_batch_start": False,
            "train_step": False,
            "on_training_batch_end": False,
            "on_training_end": False,
            "eval": False,
            "on_validation_start": False,
            "on_validation_batch_start": False,
            "valid_step": False,
            "on_validation_batch_end": False,
            "on_validation_end": False,
            "have_metrics_improved": False,
        }

        self.called_test_funs: dict[str, bool] = {
            "load_for_testing": False,
            "eval": False,
            "on_testing_start": False,
            "on_testing_batch_start": False,
            "test_step": False,
            "on_testing_batch_end": False,
            "on_testing_end": False,
        }

    def setup(self, fast_init: bool = False) -> None:
        self.called_train_funs["setup"] = True

    def load_for_testing(self) -> None:
        self.called_test_funs["load_for_testing"] = True

    def train(self) -> None:
        self.called_train_funs["train"] = True

    def on_training_start(self) -> None:
        self.called_train_funs["on_training_start"] = True

    def on_training_batch_start(self, state) -> None:
        self.called_train_funs["on_training_batch_start"] = True

    def train_step(self, batch, state) -> None:
        self.called_train_funs["train_step"] = True

    def on_training_batch_end(self, state, should_log: bool = False) -> None:
        self.called_train_funs["on_training_batch_end"] = True

    def on_training_end(self) -> None:
        self.called_train_funs["on_training_end"] = True

    def eval(self) -> None:
        self.called_train_funs["eval"] = True
        self.called_test_funs["eval"] = True

    def on_validation_start(self, validation_cycle) -> None:
        self.called_train_funs["on_validation_start"] = True

    def on_validation_batch_start(self, step) -> None:
        self.called_train_funs["on_validation_batch_start"] = True

    def valid_step(self, batch, step) -> None:
        self.called_train_funs["valid_step"] = True

    def on_validation_batch_end(self, step) -> None:
        self.called_train_funs["on_validation_batch_end"] = True

    def on_validation_end(self, cycle) -> None:
        self.called_train_funs["on_validation_end"] = True

    def have_metrics_improved(self) -> bool:
        self.called_train_funs["have_metrics_improved"] = True
        return True

    def on_testing_start(self) -> None:
        self.called_test_funs["on_testing_start"] = True

    def on_testing_batch_start(self, step) -> None:
        self.called_test_funs["on_testing_batch_start"] = True

    def test_step(self, batch, step) -> None:
        self.called_test_funs["test_step"] = True

    def on_testing_batch_end(self, step) -> None:
        self.called_test_funs["on_testing_batch_end"] = True

    def on_testing_end(self) -> None:
        self.called_test_funs["on_testing_end"] = True


class RestartModel(hlm.Model):
    def __init__(self, val_count: int = -1) -> None:
        super().__init__("test-restart")
        self.batches: list[npt.NDArray] = []
        self.val_count = val_count

    def setup(self, fast_init: bool = False) -> None:
        pass

    def on_training_batch_start(self, state: hlt.TrainingState) -> None:
        if self.val_count == state.validation_cycles:
            raise RuntimeError("stop")

    def train_step(self, batch: torch.Tensor, state) -> None:
        as_np = F.tensor_to_numpy(batch)
        self.batches.append(as_np)


class TestTrainingUnit:
    def test_from_str(self) -> None:
        assert hlt.TrainingUnit.from_str("epoch") == hlt.TrainingUnit.EPOCH
        assert hlt.TrainingUnit.from_str("iteration") == hlt.TrainingUnit.ITERATION

        with pytest.raises(ValueError):
            hlt.TrainingUnit.from_str("foo")


class TestTrainer:
    def test_find_chkpt(self, tmp_path: pathlib.Path) -> None:
        test_cases = [
            (tmp_path / "test_epoch_1_iter_10.pth", True),
            (tmp_path / "test.txt", False),
            (tmp_path / "test.pth", False),
            (tmp_path / "test_epoch_1.pth", False),
            (tmp_path / "test_iter_1.pth", False),
        ]

        for case in test_cases:
            file, exp = case
            file.touch()
            ret = hlt.find_last_checkpoint(tmp_path)
            if exp:
                assert ret is not None
            else:
                assert ret is None
            file.unlink()

    def check_trainer_flags(self, **kwargs) -> None:
        with pytest.raises(ValueError):
            hlt.Trainer(**kwargs)

    def test_trainer_validate(self, tmp_path: pathlib.Path) -> None:
        hlt.Trainer()
        invalid_root = tmp_path / "test.txt"
        invalid_root.touch()

        self.check_trainer_flags(total_steps=3.0)
        self.check_trainer_flags(chkpt_frequency=0)
        self.check_trainer_flags(print_frequency=0)
        self.check_trainer_flags(valid_frequency=0)
        self.check_trainer_flags(enable_deterministic=True, enable_cudnn_benchmark=True)
        self.check_trainer_flags(total_steps=float("inf"), early_stop_cycles=0)
        self.check_trainer_flags(enable_tensorboard=True, run_path=None)
        self.check_trainer_flags(enable_tensorboard=True, run_path=invalid_root)
        self.check_trainer_flags(enable_file_logging=True, log_path=None)
        self.check_trainer_flags(enable_file_logging=True, log_path=invalid_root)

    def test_trainer_device_flags(self) -> None:
        t = hlt.Trainer(use_cpu=True)
        assert t._use_cpu
        assert t._device == torch.device("cpu")
        assert t._map_loc == {"cuda:0": "cpu"}
        assert t._gpu_ids == []
        assert not t._is_distributed

        if torch.cuda.is_available():
            t = hlt.Trainer()
            devices = list(range(torch.cuda.device_count()))
            assert not t._use_cpu
            assert t._gpu_ids == devices
            assert t._is_distributed if len(devices) > 1 else not t._is_distributed

            with pytest.raises(ValueError):
                devs = copy.deepcopy(devices)
                devs.append(10)
                hlt.Trainer(gpus=devs)

            with pytest.raises(ValueError):
                hlt.Trainer(gpus=[len(devices)])

    def check_training_loops(
        self,
        trainer: hlt.Trainer,
        num_chkpts: int = 0,
        chkpt_root: pathlib.Path | None = None,
        fit: bool = True,
    ) -> None:
        datamodule = RandomDatamodule()
        model = CheckFunModel()

        called_funs: dict[str, bool]
        if fit:
            trainer.fit(model, datamodule)
            called_funs = model.called_train_funs
        else:
            trainer.test(model, datamodule)
            called_funs = model.called_test_funs

        for _, seen in called_funs.items():
            assert seen

        if fit:
            assert chkpt_root is not None
            assert num_chkpts == len(list(chkpt_root.glob("*.pth")))

    def test_fit_iter(self, tmp_path: pathlib.Path) -> None:
        self.check_training_loops(
            hlt.Trainer(
                train_unit=hlt.TrainingUnit.ITERATION,
                total_steps=10,
                valid_frequency=5,
                chkpt_frequency=5,
                use_cpu=True,
                enable_progress_bar=False,
                chkpt_root=tmp_path / "chkpt",
            ),
            2,
            tmp_path / "chkpt/test-model",
        )

    def test_fit_epoch(self, tmp_path: pathlib.Path) -> None:
        self.check_training_loops(
            hlt.Trainer(
                train_unit=hlt.TrainingUnit.EPOCH,
                total_steps=2,
                valid_frequency=1,
                chkpt_frequency=1,
                use_cpu=True,
                chkpt_root=tmp_path / "chkpt",
            ),
            2,
            tmp_path / "chkpt/test-model",
        )

    def test_testing(self) -> None:
        self.check_training_loops(hlt.Trainer(use_cpu=True), fit=False)

    def get_restart_trainer(
        self, unit: hlt.TrainingUnit, chkpt_root: pathlib.Path
    ) -> hlt.Trainer:
        if unit == hlt.TrainingUnit.ITERATION:
            total_steps = DATASET_SIZE
            valid_frequency = DATASET_SIZE // 2
        else:
            total_steps = 2
            valid_frequency = 1

        return hlt.Trainer(
            train_unit=unit,
            total_steps=total_steps,
            valid_frequency=valid_frequency,
            chkpt_frequency=valid_frequency,
            use_cpu=True,
            chkpt_root=chkpt_root,
        )

    def get_restart_model_and_datamodule(
        self, val_count: int = -1
    ) -> tuple[RandomDatamodule, RestartModel]:
        return RandomDatamodule(), RestartModel(val_count)

    def clear_chkpts(self, chkpt_root: pathlib.Path) -> None:
        for chkpt in chkpt_root.glob("*.pth"):
            chkpt.unlink()

    def check_batches(
        self, exp_batches: list[npt.NDArray], ret_batches: list[npt.NDArray]
    ) -> None:
        assert len(exp_batches) == len(ret_batches)
        for exp, ret in zip(exp_batches, ret_batches, strict=True):
            assert np.all(exp == ret)

    def check_restart_trainer(
        self, unit: hlt.TrainingUnit, tmp_path: pathlib.Path
    ) -> None:
        datamodule, model = self.get_restart_model_and_datamodule()
        trainer = self.get_restart_trainer(unit, tmp_path)
        chkpt_root = tmp_path / model.save_name

        trainer.fit(model, datamodule)
        batches = model.batches

        self.clear_chkpts(chkpt_root)
        del datamodule, model, trainer

        datamodule, model = self.get_restart_model_and_datamodule(1)
        trainer = self.get_restart_trainer(unit, tmp_path)
        with pytest.raises(RuntimeError):
            trainer.fit(model, datamodule)

        # Clear out everything again and restart.
        ret_batches = model.batches
        del datamodule, trainer, model

        datamodule, model = self.get_restart_model_and_datamodule()
        trainer = self.get_restart_trainer(unit, tmp_path)
        model.batches = ret_batches
        trainer.fit(model, datamodule)

        self.check_batches(batches, model.batches)

    def test_restart_iter(self, tmp_path: pathlib.Path) -> None:
        self.check_restart_trainer(hlt.TrainingUnit.ITERATION, tmp_path)

    def test_restart_epoch(self, tmp_path: pathlib.Path) -> None:
        self.check_restart_trainer(hlt.TrainingUnit.EPOCH, tmp_path)



import typing

import torch
from torch import nn
from torch.nn import init
from torch.nn.modules import batchnorm as bn

from helios import core

NETWORK_REGISTRY = core.Registry("network")


def create_network(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> nn.Module:
    """
    Create the network for the given type.

    Args:
        type_name (str): the type of the network to create.
        args: positional arguments to pass into the network.
        kwargs: keyword arguments to pass into the network.
    Returns:
        nn.Module: the network.
    """
    return NETWORK_REGISTRY.get(type_name)(*args, **kwargs)


@torch.no_grad()
def default_init_weights(
    module_list: list[nn.Module] | nn.Module,
    scale: float = 1,
    bias_fill: float = 0,
    **kwargs: typing.Any,
) -> None:
    """
    Initialize network weights.

    Args:
        module_list (list[nn.Module] | nn.Module): the list of modules to initialize.
        scale (float): scale initialized weights, especially for residual blocks.
        bias_fill (float): bias fill value.
        kwargs (Any): keyword arguments for the torch.nn.init.kaiming_normal_ function.
    """
    if not isinstance(module_list, list):
        module_list = [module_list]

    for module in module_list:
        for m in module.modules():
            if isinstance(m, nn.Conv2d | nn.Linear):
                init.kaiming_normal_(m.weight, **kwargs)
                m.weight.data *= scale
                if m.bias is not None:
                    m.bias.data.fill_(bias_fill)
            elif isinstance(m, bn._BatchNorm):  # noqa: SLF001
                init.constant_(m.weight, 1)
                if m.bias is not None:
                    m.bias.data.fill_(bias_fill)
